__all__ = ["quality","predict","tisdiff","transplot"]

from . import quality, predict, tisdiff, transplot
'''
command lines
'''

commands= {
    "quality" : quality,
    "predict" : predict,
    "tisdiff" : tisdiff,
    "transplot" : transplot
}
